#!/usr/bin/env python3

""" Recursively search and download album covers for a music library. """

import argparse
import collections
import concurrent.futures
import itertools
import logging
import multiprocessing
import os
import shutil
import sys
import time

import mutagen

import sacad


AUDIO_EXTENSIONS = ("aac",
                    "ape",
                    "flac",
                    "m4a",
                    "mp3",
                    "mp4",
                    "ogg",
                    "oga",
                    "opus",
                    "wv")


def analyze_lib(lib_dir, cover_filename):
  """ Recursively analyze library, and return a dict of path -> (artist, album). """
  work = {}
  scrobbler = itertools.cycle("|/-\\")
  stats = collections.OrderedDict(((k, 0) for k in("dirs", "files", "albums", "missing covers", "errors")))
  time_progress_shown = show_analyze_progress(stats, scrobbler)
  for rootpath, rel_dirpaths, rel_filepaths in os.walk(lib_dir):
    stats["dirs"] += 1
    metadata, time_progress_shown = analyze_dir(stats, rootpath, rel_filepaths, cover_filename, scrobbler, time_progress_shown)
    if all(metadata):
      work[rootpath] = metadata
  show_analyze_progress(stats, scrobbler, end=True)
  return work


def get_metadata(audio_filepaths):
  """ Return a tuple of album, artist from a list of audio files. """
  artist, album = None, None
  for audio_filepath in audio_filepaths:
    mf = mutagen.File(audio_filepath)
    if mf is None:
      continue
    artist = mf.get("albumartist",
                    mf.get("artist", None))
    if artist is not None:
      artist = artist[0]
    album = mf.get("_album",
                   mf.get("album",
                          None))
    if album is not None:
      album = album[0]
    if artist and album:
      # stop at the first file that succeeds (for performance)
      break
  return artist, album


def analyze_dir(stats, parent_dir, rel_filepaths, cover_filename, scrobbler, time_progress_shown):
  """ Analyze a directory (non recursively) to get its album metadata if it is one. """
  metadata = None, None
  audio_filepaths = []
  for rel_filepath in rel_filepaths:
    stats["files"] += 1
    time_progress_shown = show_analyze_progress(stats, scrobbler, time_progress_shown=time_progress_shown)
    try:
      ext = os.path.splitext(rel_filepath)[1][1:].lower()
    except IndexError:
      continue
    if ext in AUDIO_EXTENSIONS:
      audio_filepaths.append(os.path.join(parent_dir, rel_filepath))
  if audio_filepaths:
    stats["albums"] += 1
    if not os.path.isfile(os.path.join(parent_dir, cover_filename)):
      stats["missing covers"] += 1
      metadata = get_metadata(audio_filepaths)
      if not all(metadata):
        # failed to get metadata for this album
        stats["errors"] += 1
  return metadata, time_progress_shown


def show_analyze_progress(stats, scrobbler, *, time_progress_shown=0, end=False):
  """ Display analysis global progress. """
  now = time.monotonic()
  if (sys.stdout.isatty() and
     (end or (now - time_progress_shown > 0.1))):  # do not refresh display at each call (for performance)
    time_progress_shown = now
    print("Analyzing library %s | %s" % (next(scrobbler) if not end else "-",
                                         "  ".join(("%u %s" % (v, k)) for k, v in stats.items())),
          end="\r" if not end else "\n")
  return time_progress_shown


def get_covers(work, args):
  """ Get missing covers. """
  with concurrent.futures.ProcessPoolExecutor(max_workers=min(4, multiprocessing.cpu_count())) as executor:
    # post work
    futures = {}
    for path, (artist, album) in work.items():
      future = executor.submit(sacad.search_and_download,
                               album,
                               artist,
                               args.format,
                               args.size,
                               args.size_tolerance_prct,
                               args.amazon_tlds,
                               args.no_lq_sources,
                               os.path.join(path, args.filename))
      futures[future] = (path, artist, album)

    # follow progress
    stats = collections.OrderedDict(((k, 0) for k in("ok", "errors", "no result found")))
    errors = []
    not_found = []
    for i, future in enumerate(concurrent.futures.as_completed(futures)):
      path, artist, album = futures[future]
      try:
        status = future.result()
      except Exception:
        stats["errors"] += 1
        errors.append((path, artist, album))
      else:
        if status:
          stats["ok"] += 1
        else:
          stats["no result found"] += 1
          not_found.append((path, artist, album))
      show_get_covers_progress(i,
                               len(work),
                               stats,
                               artist=artist,
                               album=album)
  if work:
    show_get_covers_progress(i + 1, len(work), stats, end=True)
  for path, artist, album in not_found:
    print("Unable to find cover for '%s' by '%s' from '%s'" % (album, artist, path))
  for path, artist, album in errors:
    print("Error occured while searching cover for '%s' by '%s' from '%s'" % (album, artist, path))
  if errors:
    print("Please report this at https://github.com/desbma/sacad/issues")


def show_get_covers_progress(current_idx, total_count, stats, *, artist=None, album=None, end=False):
  """ Display search and download global progress. """
  if not sys.stdout.isatty():
    return
  line_width = shutil.get_terminal_size(fallback=(80, 0))[0] - 1
  print(" " * line_width, end="\r")
  print("Searching and downloading covers %u%% (%u/%u)%s | %s" % (100 * current_idx // total_count,
                                                                  current_idx,
                                                                  total_count,
                                                                  (" | Current album: '%s' '%s'" % (artist, album)) if not end else "",
                                                                  "  ".join(("%u %s" % (v, k)) for k, v in stats.items())),
        end="\r" if not end else "\n")


def cl_main():
  # parse args
  arg_parser = argparse.ArgumentParser(description="SACAD (recursive tool) v%s.%s" % (sacad.__version__,
                                                                                      __doc__),
                                       formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  arg_parser.add_argument("lib_dir",
                          help="Music library directory to recursively analyze")
  arg_parser.add_argument("size",
                          type=int,
                          help="Target image size")
  arg_parser.add_argument("filename",
                          help="Cover image filename")
  sacad.setup_common_args(arg_parser)
  args = arg_parser.parse_args()
  args.format = os.path.splitext(args.filename)[1][1:].lower()
  try:
    args.format = sacad.SUPPORTED_IMG_FORMATS[args.format]
  except KeyError:
    print("Unable to guess image format from extension, or unknown format: %s" % (args.format))
    exit(1)

  # silence the logger
  logging.basicConfig(format="%(process)d %(threadName)s: %(message)s", level=logging.ERROR)

  # do the job
  work = analyze_lib(args.lib_dir, args.filename)
  get_covers(work, args)



#!/usr/bin/env python3

import logging
import socket
import unittest
import unittest.mock
import urllib.parse

import PIL.Image
import requests

import sacad
from . import rate_watcher_test
from . import recurse_test


sacad.web_cache.DISABLE_PERSISTENT_CACHING = True


def is_internet_reachable():
  try:
    # open TCP socket to Google DNS server
    with socket.create_connection(("8.8.8.8", 53)):
      pass
  except OSError as e:
    if e.errno == 101:
      return False
    raise
  return True


def download(url, filepath):
  response = requests.get(url, timeout=5, verify=False, stream=True)
  response.raise_for_status()
  with open(filepath, "wb") as f:
    for chunk in response.iter_content(2 ** 14):
      f.write(chunk)


@unittest.skipUnless(is_internet_reachable(), "Need Internet access")
class TestSacad(unittest.TestCase):

  @staticmethod
  def getImgInfo(img_filepath):
    with open(img_filepath, "rb") as img_file:
      img = PIL.Image.open(img_file)
      format = img.format.lower()
      format = sacad.SUPPORTED_IMG_FORMATS[format]
      width, height = img.size
    return format, width, height

  def test_getMasterOfPuppetsCover(self):
    """ Search and download cover for 'Master of Puppets' with different parameters. """
    for format in sacad.cover.CoverImageFormat:
      for size in (300, 600, 1200):
        for size_tolerance in (0, 25, 50):
          with sacad.mkstemp_ctx.mkstemp(prefix="sacad_test_",
                                         suffix=".%s" % (format.name.lower())) as tmp_filepath:
            sacad.search_and_download("Master of Puppets",
                                      "Metallica",
                                      format,
                                      size,
                                      size_tolerance,
                                      (),
                                      False,
                                      tmp_filepath)
            out_format, out_width, out_height = __class__.getImgInfo(tmp_filepath)
            self.assertEqual(out_format, format)
            self.assertLessEqual(out_width, size * (100 + size_tolerance) / 100)
            self.assertGreaterEqual(out_width, size * (100 - size_tolerance) / 100)
            self.assertLessEqual(out_height, size * (100 + size_tolerance) / 100)
            self.assertGreaterEqual(out_height, size * (100 - size_tolerance) / 100)

  def test_getImageUrlMetadata(self):
    """ Download the beginning of image files to guess their format and resolution. """
    refs = {"https://www.nuclearblast.de/static/articles/152/152118.jpg/1000x1000.jpg": (sacad.cover.CoverImageFormat.JPEG,
                                                                                         (700, 700),
                                                                                         5),
            "http://img2-ak.lst.fm/i/u/55ad95c53e6043e3b150ba8a0a3b20a1.png": (sacad.cover.CoverImageFormat.PNG,
                                                                               (600, 600),
                                                                               1)}
    for url, (ref_fmt, ref_size, block_read) in refs.items():
      sacad.CoverSourceResult.getImageMetadata = unittest.mock.Mock(wraps=sacad.CoverSourceResult.getImageMetadata)
      source = unittest.mock.Mock()
      source.http_session = sacad.http.session()
      cover = sacad.CoverSourceResult(url,
                                      None,
                                      None,
                                      source=source,
                                      thumbnail_url=None,
                                      source_quality=None,
                                      check_metadata=sacad.cover.CoverImageMetadata.ALL)
      cover.updateImageMetadata()
      self.assertEqual(cover.size, ref_size)
      self.assertEqual(cover.format, ref_fmt)
      self.assertGreaterEqual(sacad.CoverSourceResult.getImageMetadata.call_count, 0)
      self.assertLessEqual(sacad.CoverSourceResult.getImageMetadata.call_count, block_read)

  def test_compareImageSignatures(self):
    """ Compare images using their signatures. """
    urls = ("http://wac.450f.edgecastcdn.net/80450F/kool1017.com/files/2013/09/cover_highway_to_hell_500x500.jpg",
            "http://www.jesus-is-savior.com/Evils%20in%20America/Rock-n-Roll/highway_to_hell-large.jpg",
            "http://i158.photobucket.com/albums/t113/gatershanks/Glee%20Alternative%20Song%20Covers/1x14%20Hell%20O/1x14Hell-O-HighwayToHell.jpg")
    with sacad.mkstemp_ctx.mkstemp(suffix="jpg") as temp_filepath1, \
            sacad.mkstemp_ctx.mkstemp(suffix="jpg") as temp_filepath2, \
            sacad.mkstemp_ctx.mkstemp(suffix="jpg") as temp_filepath3:
      img_sig = {}
      for i, (url, filepath) in enumerate(zip(urls, (temp_filepath1, temp_filepath2, temp_filepath3))):
        download(url, filepath)
        with open(filepath, "rb") as img_file:
          img_data = img_file.read()
          img_sig[i] = sacad.CoverSourceResult.computeImgSignature(img_data)
      self.assertTrue(sacad.CoverSourceResult.areImageSigsSimilar(img_sig[0], img_sig[1]))
      self.assertTrue(sacad.CoverSourceResult.areImageSigsSimilar(img_sig[1], img_sig[0]))
      self.assertFalse(sacad.CoverSourceResult.areImageSigsSimilar(img_sig[0], img_sig[2]))
      self.assertFalse(sacad.CoverSourceResult.areImageSigsSimilar(img_sig[1], img_sig[2]))
      self.assertFalse(sacad.CoverSourceResult.areImageSigsSimilar(img_sig[2], img_sig[0]))
      self.assertFalse(sacad.CoverSourceResult.areImageSigsSimilar(img_sig[2], img_sig[1]))

  def test_coverSources(self):
    """ Check all sources return valid results with different parameters. """
    for size in range(300, 1200 + 1, 300):
      source_args = (size, 0)
      sources = [sacad.sources.LastFmCoverSource(*source_args),
                 sacad.sources.GoogleImagesWebScrapeCoverSource(*source_args),
                 sacad.sources.AmazonDigitalCoverSource(*source_args)]
      sources.extend(sacad.sources.AmazonCdCoverSource(*source_args, tld=tld) for tld in sacad.sources.AmazonCdCoverSource.TLDS)
      for source in sources:
        for artist, album in zip(("Michael Jackson", "Björk"), ("Thriller", "Vespertine")):
          results = source.search(album, artist)
          results = sacad.CoverSourceResult.preProcessForComparison(results, size, 0)
          if not (((size > 500) and isinstance(source, sacad.sources.AmazonCdCoverSource)) or
                  ((size > 600) and isinstance(source, sacad.sources.LastFmCoverSource)) or
                  (isinstance(source, sacad.sources.AmazonCdCoverSource) and (artist == "Björk") and
                   (urllib.parse.urlsplit(source.base_url).netloc.rsplit(".", 1)[-1] == "cn"))):
            self.assertGreaterEqual(len(results), 1, "%s %s %s %u" % (source.__class__.__name__,
                                                                      artist,
                                                                      album,
                                                                      size))
          for result in results:
            self.assertTrue(result.urls)
            self.assertIn(result.format, sacad.cover.CoverImageFormat)
            self.assertGreaterEqual(result.size[0], size)

    # test for specific cover not available on amazon.com, but on amazon.de
    size = 290
    source = sacad.sources.AmazonCdCoverSource(size, 0, tld="de")
    results = source.search("Dream Dance 5", "Various")
    self.assertGreaterEqual(len(results), 1)
    for result in results:
      self.assertTrue(result.urls)
      self.assertIn(result.format, sacad.cover.CoverImageFormat)
      self.assertGreaterEqual(result.size[0], size)

  def test_unaccentuate(self):
    self.assertEqual(sacad.sources.base.CoverSource.unaccentuate("EéeAàaOöoIïi"), "EeeAaaOooIii")


if __name__ == "__main__":
  # logging
  #logging.getLogger().setLevel(logging.DEBUG)
  logging.getLogger().setLevel(logging.CRITICAL + 1)

  # run tests
  unittest.main()



""" This module provides a class with a context manager to help avoid overloading web servers. """

import collections
import logging
import os
import sqlite3
import threading
import time
import urllib.parse

import lockfile


class AccessRateWatcher:

  """ Access rate limiter, supporting concurrent access by threads and/or processes. """

  thread_locks = collections.defaultdict(threading.Lock)
  thread_dict_lock = threading.Lock()

  def __init__(self, db_filepath, url, min_delay_between_accesses):
    self.domain = urllib.parse.urlsplit(url).netloc
    self.min_delay_between_accesses = min_delay_between_accesses
    self.connexion = sqlite3.connect(db_filepath)
    with self.connexion:
      self.connexion.executescript("""PRAGMA journal_mode = MEMORY;
                                      PRAGMA synchronous = OFF;
                                      CREATE TABLE IF NOT EXISTS access_timestamp (domain TEXT PRIMARY KEY,
                                                                                   timestamp FLOAT NOT NULL);""")
    self.lock_dir = os.path.join(os.path.dirname(db_filepath), "plocks")
    os.makedirs(self.lock_dir, exist_ok=True)

  def __enter__(self):
    self._waitAccess()

  def __exit__(self, exc_type, exc_value, traceback):
    self._access()
    self._releaseLock()

  def _access(self):
    """ Notify the watcher that the server is accessed. """
    with self.connexion:
      self.connexion.execute("""INSERT OR REPLACE INTO access_timestamp
                               (domain, timestamp) VALUES (?, ?)""",
                             (self.domain, time.time(),))

  def _waitAccess(self):
    """ Wait the needed time before sending a request to honor rate limit. """
    while True:
      with self.connexion:
        last_access_time = self.connexion.execute("""SELECT timestamp
                                                     FROM access_timestamp
                                                     WHERE domain = ?;""",
                                                  (self.domain,)).fetchone()
      if last_access_time is not None:
        last_access_time = last_access_time[0]
        now = time.time()
        time_since_last_access = now - last_access_time
        if time_since_last_access < self.min_delay_between_accesses:
          time_to_wait = self.min_delay_between_accesses - time_since_last_access
          logging.getLogger().debug("Sleeping for %.2fms because of rate limit" % (time_to_wait * 1000))
          time.sleep(time_to_wait)

      if self._getLock():
        break
      else:
        time.sleep(0.001)

  def _getLock(self):
    with __class__.thread_dict_lock:
      tlock = __class__.thread_locks[self.domain]
    if tlock.acquire(blocking=False):
      plock = lockfile.FileLock(os.path.join(self.lock_dir, self.domain))
      try:
        plock.acquire(timeout=0)
      except (lockfile.LockTimeout, lockfile.AlreadyLocked):
        tlock.release()
      except:
        tlock.release()
        raise
      else:
        return True
    return False

  def _releaseLock(self):
    __class__.thread_locks[self.domain].release()
    lockfile.FileLock(os.path.join(self.lock_dir, self.domain)).release()



#!/usr/bin/env python3

import collections
import logging
import os
import pickle
import random
import string
import sys
import time
import unittest

import sacad.mkstemp_ctx as mkstemp_ctx
import sacad.web_cache as web_cache


web_cache.DISABLE_PERSISTENT_CACHING = True

INFINITY = sys.maxsize


def get_random_string(length, chars=string.ascii_letters + string.digits):
  return "".join(random.choice(chars) for _ in range(length))


class TestWebCache(unittest.TestCase):

  def test_getSetDelete(self):
    """ Get/set/delete cache items using all cache parameter combinations. """
    for cache_class in (web_cache.WebCache, web_cache.ThreadedWebCache):
      for compression in web_cache.Compression:
        for compression_level in range(1, 9):
          for caching_strategy in web_cache.CachingStrategy:
            for expiration in (None, 0, INFINITY):
              for sql_crash_safe in (True, False):
                table_name = get_random_string(16, string.ascii_letters)
                with mkstemp_ctx.mkstemp(suffix=".sqlite") as cache_filepath:
                  # init cache
                  cache = cache_class(cache_filepath,
                                      table_name,
                                      caching_strategy=caching_strategy,
                                      expiration=expiration,
                                      compression=compression,
                                      compression_level=compression_level,
                                      safe_mode=sql_crash_safe)
                  already_used_keys = set()
                  item_count = 0

                  for req_type in ("get", "post"):
                    for item_count in range(item_count + 1, item_count + 4):
                      while True:
                        # generate cache key
                        key = get_random_string(16)
                        if req_type == "post":
                          key = key, collections.OrderedDict(((k, v) for k, v in zip((get_random_string(8) for _ in range(4)),
                                                                                     (get_random_string(16) for _ in range(4)))))

                        # ensure key is unique for this cache
                        bin_key = pickle.dumps(key)
                        if bin_key not in already_used_keys:
                          already_used_keys.add(bin_key)
                          break

                      # generate cache data
                      data = os.urandom(2 ** 13)

                      # check cache size
                      self.assertEqual(len(cache), item_count - 1)

                      # check key is not in cache
                      self.assertNotIn(key, cache)
                      with self.assertRaises(KeyError):
                        cache[key]
                      with self.assertRaises(KeyError):
                        del cache[key]

                      # add data to cache
                      cache[key] = data

                      # check key is in cache
                      self.assertIn(key, cache)
                      self.assertEqual(cache[key], data)

                      # check cache size
                      self.assertEqual(len(cache), item_count)

                      # delete cache item
                      del cache[key]

                      # check it is not in cache anymore
                      self.assertNotIn(key, cache)
                      with self.assertRaises(KeyError):
                        cache[key]
                      with self.assertRaises(KeyError):
                        del cache[key]

                      # check cache size
                      self.assertEqual(len(cache), item_count - 1)

                      # check other keys are still here
                      for old_key in map(pickle.loads, already_used_keys):
                        if old_key != key:
                          self.assertIn(old_key, cache)

                      # add cache item again
                      cache[key] = data

  def test_getCacheHitStats(self):
    """ Get cache stats using all cache parameter combinations. """
    for cache_class in (web_cache.WebCache, web_cache.ThreadedWebCache):
      for compression in web_cache.Compression:
        for compression_level in range(1, 9):
          for caching_strategy in web_cache.CachingStrategy:
            for expiration in (None, 0, INFINITY):
              for sql_crash_safe in (True, False):
                table_name = get_random_string(16, string.ascii_letters)
                with mkstemp_ctx.mkstemp(suffix=".sqlite") as cache_filepath:
                  # init cache
                  cache = cache_class(cache_filepath,
                                      table_name,
                                      caching_strategy=caching_strategy,
                                      expiration=expiration,
                                      compression=compression,
                                      compression_level=compression_level,
                                      safe_mode=sql_crash_safe)

                  i = 0
                  for req_type in ("get", "post"):
                    for i in range(i + 1, 5):
                      # generate item
                      key = "%s_%u" % (req_type, i)
                      if req_type == "post":
                        key = key, collections.OrderedDict(((k, v) for k, v in zip((get_random_string(4) for _ in range(2)),
                                                                                   (get_random_string(8) for _ in range(2)))))
                      data = os.urandom(2 ** 13)

                      # add item
                      cache[key] = data

                      # check cache hit stats
                      self.assertEqual(cache.getCacheHitStats(), (i - 1, i - 1))
                      self.assertIn(key, cache)
                      self.assertEqual(cache.getCacheHitStats(), (i, i - 1))
                      self.assertNotIn("(o_o)", cache)
                      self.assertEqual(cache.getCacheHitStats(), (i, i))

  def test_purge(self):
    """ Purge obsolete cache entries. """
    for cache_class in (web_cache.WebCache, web_cache.ThreadedWebCache):
      for caching_strategy in web_cache.CachingStrategy:
        for expiration in (None, 2, INFINITY):
          table_name = get_random_string(16, string.ascii_letters)
          with mkstemp_ctx.mkstemp(suffix=".sqlite") as cache_filepath:
            # init cache
            cache = cache_class(cache_filepath,
                                table_name,
                                caching_strategy=caching_strategy,
                                expiration=expiration)

            # add items
            for req_type in ("get", "post"):
              for i in range(5):
                key = "%s_%u" % (req_type, i)
                if req_type == "post":
                  key = key, collections.OrderedDict(((k, v) for k, v in zip((get_random_string(4) for _ in range(2)),
                                                                             (get_random_string(8) for _ in range(2)))))
                data = os.urandom(2 ** 13)
                cache[key] = data

            # purge
            purged_count = cache.purge()
            if expiration and (expiration != INFINITY):
              # before expiration, nothing should have been purged
              time.sleep(1)
              self.assertEqual(purged_count, 0)
              self.assertEqual(len(cache), 10)
              # wait for expiration
              time.sleep(expiration)
              # after expiration, all should have been purged
              purged_count = cache.purge()
              self.assertEqual(purged_count, 10)
              self.assertEqual(len(cache), 0)
            else:
              # nothing should have been purged
              self.assertEqual(purged_count, 0)
              self.assertEqual(len(cache), 10)


if __name__ == "__main__":
  # disable logging
  logging.basicConfig(level=logging.CRITICAL + 1)

  # run tests
  unittest.main()



import collections
import urllib.parse

import lxml.cssselect
import lxml.etree

from sacad.cover import CoverImageFormat, CoverImageMetadata, CoverSourceQuality, CoverSourceResult
from sacad.sources.base import CoverSource


class AmazonCdCoverSourceResult(CoverSourceResult):

  def __init__(self, *args, **kwargs):
    super().__init__(*args, source_quality=CoverSourceQuality.NORMAL, **kwargs)


class AmazonCdCoverSource(CoverSource):

  """ Cover source returning Amazon.com audio CD images. """

  TLDS = ("com", "ca", "cn", "fr", "de", "co.jp", "co.uk")

  def __init__(self, *args, tld="com", **kwargs):
    assert(tld in __class__.TLDS)
    self.base_url = "https://www.amazon.%s/gp/search" % (tld)
    super().__init__(*args, allow_cookies=True, **kwargs)

  def getSearchUrl(self, album, artist):
    """ See CoverSource.getSearchUrl. """
    params = collections.OrderedDict()
    params["search-alias"] = "popular"
    params["field-artist"] = __class__.unaccentuate(artist.lower())
    params["field-title"] = __class__.unaccentuate(album.lower())
    params["sort"] = "relevancerank"
    return __class__.assembleUrl(self.base_url, params)

  def parseResults(self, api_data):
    """ See CoverSource.parseResults. """
    results = []

    # parse page
    parser = lxml.etree.HTMLParser()
    html = lxml.etree.XML(api_data.decode("utf-8"), parser)
    results_selector = lxml.cssselect.CSSSelector("#atfResults li.s-result-item")
    img_selector = lxml.cssselect.CSSSelector("img.s-access-image")
    product_link_selector = lxml.cssselect.CSSSelector("a.s-access-detail-page")
    product_page_img_selector = lxml.cssselect.CSSSelector("img#landingImage")
    result_divs = results_selector(html)
    for rank, result_div in enumerate(result_divs, 1):
      try:
        img_node = img_selector(result_div)[0]
      except IndexError:
        # no image for that product
        continue
      # get thumbnail & full image url
      thumbnail_url = img_node.get("src")
      url_parts = thumbnail_url.rsplit(".", 2)
      img_url = ".".join((url_parts[0], url_parts[2]))
      # assume size is fixed
      size = (500, 500)
      check_metadata = CoverImageMetadata.SIZE
      # try to get higher res image...
      if ((self.target_size > size[0]) and  # ...only if needed
              (rank <= 3)):  # and only for first 3 results because this is time
                             # consuming (1 more GET request per result)
        product_url = product_link_selector(result_div)[0].get("href")
        product_url = urllib.parse.urlsplit(product_url)
        product_url_query = urllib.parse.parse_qsl(product_url.query)
        product_url_query = collections.OrderedDict(product_url_query)
        del product_url_query["qid"]  # remove timestamp from url to improve future cache hit rate
        product_url_query = urllib.parse.urlencode(product_url_query)
        product_url_no_ts = urllib.parse.urlunsplit(product_url[:3] + (product_url_query,) + product_url[4:])
        cache_hit, product_page_data = self.fetchResults(product_url_no_ts)
        product_page_html = lxml.etree.XML(product_page_data.decode("latin-1"), parser)
        try:
          img_node = product_page_img_selector(product_page_html)[0]
        except IndexError:
          # unable to get better image
          pass
        else:
          better_img_url = img_node.get("data-old-hires")
          # img_node.get("data-a-dynamic-image") contains json with image urls too, but they are not larger than
          # previous 500px image and are often covered by autorip badges (can be removed by cleaning url though)
          if better_img_url:
            img_url = better_img_url
            size_url_hint = img_url.rsplit(".", 2)[1].strip("_")
            assert(size_url_hint.startswith("SL"))
            size_url_hint = int(size_url_hint[2:])
            size = (size_url_hint, size_url_hint)
            check_metadata = CoverImageMetadata.NONE
        if not cache_hit:
          CoverSource.api_cache[product_url_no_ts] = product_page_data

      # assume format is always jpg
      format = CoverImageFormat.JPEG
      # add result
      results.append(AmazonCdCoverSourceResult(img_url,
                                               size,
                                               format,
                                               thumbnail_url=thumbnail_url,
                                               source=self,
                                               rank=rank,
                                               check_metadata=check_metadata))

    return results



""" Common HTTP code. """

import contextlib
import http.cookiejar
import logging
import os
import socket

import redo
import requests


IS_TRAVIS = os.getenv("CI") and os.getenv("TRAVIS")
HTTP_NORMAL_TIMEOUT_S = 30.1 if IS_TRAVIS else 9.1
HTTP_SHORT_TIMEOUT_S = 18.1 if IS_TRAVIS else 3.1
HTTP_MAX_ATTEMPTS = 20 if IS_TRAVIS else 3
DEFAULT_USER_AGENT = "Mozilla/5.0"


def query(url, *, session, watcher=None, post_data=None, headers=None, verify=True):
  """ Send a GET/POST request, retry if it fails, and return response content. """
  if headers is None:
    headers = {}
  if "User-Agent" not in headers:
    headers["User-Agent"] = DEFAULT_USER_AGENT

  for attempt, _ in enumerate(redo.retrier(attempts=HTTP_MAX_ATTEMPTS,
                                           sleeptime=1.5,
                                           max_sleeptime=5,
                                           sleepscale=1.25,
                                           jitter=1),
                              1):
    try:
      with contextlib.ExitStack() as context_manager:
        if watcher is not None:
          context_manager.enter_context(watcher)
        if post_data is not None:
          response = session.post(url,
                                  data=post_data,
                                  headers=headers,
                                  timeout=HTTP_NORMAL_TIMEOUT_S,
                                  verify=verify)
        else:
          response = session.get(url,
                                 headers=headers,
                                 timeout=HTTP_NORMAL_TIMEOUT_S,
                                 verify=verify)
      break

    except requests.exceptions.SSLError:
      raise

    except (socket.timeout, requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
      logging.getLogger().warning("Querying '%s' failed (attempt %u/%u): %s %s" % (url,
                                                                                   attempt,
                                                                                   HTTP_MAX_ATTEMPTS,
                                                                                   e.__class__.__qualname__,
                                                                                   e))
      if attempt == HTTP_MAX_ATTEMPTS:
        raise

  response.raise_for_status()

  return response.content


def is_reachable(url, *, session, watcher=None, headers=None, verify=True, response_headers=None):
  """ Send a HEAD request with short timeout, return True if ressource has 2xx status code, False instead. """
  if headers is None:
    headers = {}
  if "User-Agent" not in headers:
    headers["User-Agent"] = DEFAULT_USER_AGENT

  try:
    for attempt, _ in enumerate(redo.retrier(attempts=HTTP_MAX_ATTEMPTS,
                                             sleeptime=1.5,
                                             max_sleeptime=3,
                                             sleepscale=1.25,
                                             jitter=1),
                                1):
      try:
        with contextlib.ExitStack() as context_manager:
          if watcher is not None:
            context_manager.enter_context(watcher)
          response = session.head(url,
                                  headers=headers,
                                  timeout=HTTP_SHORT_TIMEOUT_S,
                                  verify=verify)
          break

      except (socket.timeout, requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
        logging.getLogger().warning("Probing '%s' failed (attempt %u/%u): %s %s" % (url,
                                                                                    attempt,
                                                                                    HTTP_MAX_ATTEMPTS,
                                                                                    e.__class__.__qualname__,
                                                                                    e))
        if attempt == HTTP_MAX_ATTEMPTS:
          return False

    response.raise_for_status()

    if response_headers is not None:
      response_headers.update(response.headers)

  except requests.exceptions.HTTPError as e:
    logging.getLogger().warning("Probing '%s' failed: %s %s" % (url, e.__class__.__qualname__, e))
    return False

  return True


def fast_streamed_query(url, *, session, headers=None, verify=True):
  """ Send a GET request with short timeout, do not retry, and return streamed response. """
  if headers is None:
    headers = {}
  if "User-Agent" not in headers:
    headers["User-Agent"] = DEFAULT_USER_AGENT

  response = session.get(url,
                         headers=headers,
                         timeout=HTTP_SHORT_TIMEOUT_S,
                         verify=verify,
                         stream=True)

  response.raise_for_status()

  return response


def session(allow_cookies=False):
  """ Return a HTTP session to use to benefit from TCP connection reuse. It also optionally refuses cookies. """
  s = requests.Session()
  if not allow_cookies:
    cp = http.cookiejar.DefaultCookiePolicy(allowed_domains=[])
    s.cookies.set_policy(cp)
  return s


# silence third party module loggers
logging.getLogger("redo").setLevel(logging.ERROR)
logging.getLogger("requests").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)
try:
  requests.packages.urllib3.disable_warnings()
except:
  pass



#!/usr/bin/env python3

import logging
import os
import tempfile
import time
import unittest

from sacad.rate_watcher import AccessRateWatcher


class TestRateWatcher(unittest.TestCase):

  def test_minDelayBetweenAccesses(self):
    with tempfile.TemporaryDirectory() as tmp_dir:
      db_filepath = os.path.join(tmp_dir, "db.sqlite")
      with AccessRateWatcher(db_filepath,
                             "http://1.domain.com/abcd",
                             min_delay_between_accesses=1):
        time_first_access_domain1 = time.time()
      with AccessRateWatcher(db_filepath,
                             "http://2.domain.com/abcd",
                             min_delay_between_accesses=1):
        time_first_access_domain2 = time.time()
      self.assertLess(time_first_access_domain2 - time_first_access_domain1, 1)
      with AccessRateWatcher(db_filepath,
                             "http://1.domain.com/efgh",
                             min_delay_between_accesses=1):
        time_second_access_domain1 = time.time()
      self.assertGreater(time_second_access_domain1 - time_first_access_domain1, 1)
      with AccessRateWatcher(db_filepath,
                             "http://2.domain.com/efgh",
                             min_delay_between_accesses=1):
        time_second_access_domain2 = time.time()
      self.assertLess(time_second_access_domain2 - time_second_access_domain1, 1)
      with AccessRateWatcher(db_filepath,
                             "http://2.domain.com/ijkl",
                             min_delay_between_accesses=2):
        time_third_access_domain2 = time.time()
      self.assertGreater(time_third_access_domain2 - time_second_access_domain2, 2)


if __name__ == "__main__":
  # disable logging
  logging.basicConfig(level=logging.CRITICAL + 1)

  # run tests
  unittest.main()



import collections
import logging
import urllib.parse

import lxml.cssselect
import lxml.etree

from sacad.cover import CoverImageFormat, CoverImageMetadata, CoverSourceQuality, CoverSourceResult
from sacad.sources.base import CoverSource


class AmazonDigitalCoverSourceResult(CoverSourceResult):

  def __init__(self, *args, **kwargs):
    super().__init__(*args, source_quality=CoverSourceQuality.NORMAL, **kwargs)


class AmazonDigitalCoverSource(CoverSource):

  """ Cover source returning Amazon.com digital music images. """

  BASE_URL = "https://www.amazon.com/gp"
  DYNAPI_KEY = "A17SFUTIVB227Z"

  def __init__(self, *args, **kwargs):
    super().__init__(*args, allow_cookies=True, **kwargs)

  def getSearchUrl(self, album, artist):
    """ See CoverSource.getSearchUrl. """
    url = "%s/search" % (__class__.BASE_URL)
    params = collections.OrderedDict()
    params["search-alias"] = "digital-music"
    params["field-keywords"] = " ".join(map(__class__.unaccentuate,
                                            map(str.lower,
                                                (artist, album))))
    params["sort"] = "relevancerank"
    return __class__.assembleUrl(url, params)

  def parseResults(self, api_data):
    """ See CoverSource.parseResults. """
    results = []

    # parse page
    parser = lxml.etree.HTMLParser()
    html = lxml.etree.XML(api_data.decode("utf-8"), parser)
    results_selectors = (lxml.cssselect.CSSSelector("div#dm_mp3Player div.mp3Cell"),
                         lxml.cssselect.CSSSelector("div#dm_mp3Player li.s-mp3-federated-bar-item"))
    img_selectors = (lxml.cssselect.CSSSelector("img.productImage"),
                     lxml.cssselect.CSSSelector("img.s-access-image"))
    link_selector = lxml.cssselect.CSSSelector("a")
    slice_count_to_res = {1: 600, 2: 700, 3: 1050, 4: 1400}

    for page_struct_version in range(len(results_selectors)):
      result_divs = results_selectors[page_struct_version](html)
      if result_divs:
        break

    for rank, result_div in enumerate(result_divs, 1):
      # get thumbnail & full image url
      img_node = img_selectors[page_struct_version](result_div)[0]
      thumbnail_url = img_node.get("src")
      thumbnail_url = thumbnail_url.replace("Stripe-Prime-Only", "")
      url_parts = thumbnail_url.rsplit(".", 2)
      img_url = ".".join((url_parts[0], url_parts[2]))

      # assume size is fixed
      size = (500, 500)

      # try to get higher res image...
      if self.target_size > size[0]:  # ...but only if needed
        logging.getLogger().debug("Looking for optimal subimages configuration...")
        product_url = link_selector(result_div)[0].get("href")
        product_url = urllib.parse.urlsplit(product_url)
        product_id = product_url.path.split("/")[3]

        found = False
        for slice_count in range(4, 1, -1):
          for square_sub_img in (True, False):
            if square_sub_img and (slice_count == 4):
              # unfortunately, this one is never available
              continue

            logging.getLogger().debug("Trying %u %ssquare subimages..." % (slice_count ** 2,
                                                                           "non " if not square_sub_img else ""))
            urls = tuple(self.generateImgUrls(product_id, __class__.DYNAPI_KEY, slice_count, square_sub_img))
            url_ok = self.probeUrl(urls[-1])
            if not url_ok:
              # images at this size are not available
              continue

            # images at this size are available
            found = True
            break

          if found:
            break

        if found:
          img_url = urls
          size = (slice_count_to_res[slice_count],) * 2

      # assume format is always jpg
      format = CoverImageFormat.JPEG

      # add result
      results.append(AmazonDigitalCoverSourceResult(img_url,
                                                    size,
                                                    format,
                                                    thumbnail_url=thumbnail_url,
                                                    source=self,
                                                    rank=rank,
                                                    check_metadata=CoverImageMetadata.SIZE))

    return results

  def generateImgUrls(self, product_id, dynapi_key, slice_count, square_sub_img):
    """ Generate URLs for slice_count^2 subimages of a product. """
    for x in range(slice_count):
      for y in range(slice_count):
        yield ("http://z2-ec2.images-amazon.com/R/1/a=" + product_id +
               "+c=" + dynapi_key +
               "+d=_SCR%28" + str(slice_count - 1 + int(square_sub_img)) + "," + str(x) + "," + str(y) + "%29_=.jpg")



from sacad.sources.amazoncd import AmazonCdCoverSource, AmazonCdCoverSourceResult
from sacad.sources.amazondigital import AmazonDigitalCoverSource, AmazonDigitalCoverSourceResult
from sacad.sources.lastfm import LastFmCoverSource, LastFmCoverSourceResult
from sacad.sources.google_images import GoogleImagesWebScrapeCoverSource, GoogleImagesCoverSourceResult



#!/usr/bin/env python3

""" Smart Automatic Cover Art Downloader : search and download music album covers. """

__version__ = "1.6.10"
__author__ = "desbma"
__license__ = "MPL 2.0"

import argparse
import functools
import logging
import os
import sys

from sacad import colored_logging
from sacad import sources
from sacad.cover import CoverSourceResult, HAS_JPEGOPTIM, HAS_OPTIPNG, SUPPORTED_IMG_FORMATS


def search_and_download(album, artist, format, size, size_tolerance_prct, amazon_tlds, no_lq_sources, out_filepath):
  """ Search and download a cover, return True if success, False instead. """
  # display warning if optipng or jpegoptim are missing
  if not HAS_JPEGOPTIM:
    logging.getLogger().warning("jpegoptim could not be found, JPEG crunching will be disabled")
  if not HAS_OPTIPNG:
    logging.getLogger().warning("optipng could not be found, PNG crunching will be disabled")

  # register sources
  source_args = (size, size_tolerance_prct)
  cover_sources = [sources.LastFmCoverSource(*source_args),
                   sources.AmazonCdCoverSource(*source_args),
                   sources.AmazonDigitalCoverSource(*source_args)]
  for tld in amazon_tlds:
    cover_sources.append(sources.AmazonCdCoverSource(*source_args, tld=tld))
  if not no_lq_sources:
    cover_sources.append(sources.GoogleImagesWebScrapeCoverSource(*source_args))

  # search
  results = []
  for cover_source in cover_sources:  # TODO to that in processes
    results.extend(cover_source.search(album, artist))

  # sort results
  results = CoverSourceResult.preProcessForComparison(results, size, size_tolerance_prct)
  results.sort(reverse=True,
               key=functools.cmp_to_key(functools.partial(CoverSourceResult.compare,
                                                          target_size=size,
                                                          size_tolerance_prct=size_tolerance_prct)))
  if not results:
    logging.getLogger().info("No results")

  # download
  for result in results:
    try:
      result.get(format, size, size_tolerance_prct, out_filepath)
    except Exception as e:
      logging.getLogger().warning("Download of %s failed: %s %s" % (result,
                                                                    e.__class__.__qualname__,
                                                                    e))
      continue
    else:
      return True

  return False


def setup_common_args(arg_parser):
  arg_parser.add_argument("-t",
                          "--size-tolerance",
                          type=int,
                          default=25,
                          dest="size_tolerance_prct",
                          help="""Tolerate this percentage of size difference with the target size.
                                  Note that covers with size above or close to the target size will still be preferred
                                  if available""")
  arg_parser.add_argument("-a",
                          "--amazon-sites",
                          nargs="+",
                          choices=sources.AmazonCdCoverSource.TLDS[1:],
                          default=(),
                          dest="amazon_tlds",
                          help="""Amazon site TLDs to use as search source, in addition to amazon.com""")
  arg_parser.add_argument("-d",
                          "--disable-low-quality-sources",
                          action="store_true",
                          default=False,
                          dest="no_lq_sources",
                          help="""Disable cover sources that may return unreliable results (ie. Google Images).
                                  It will speed up search and improve reliability, but may fail to find results for
                                  some difficult searches.""")


def cl_main():
  # parse args
  arg_parser = argparse.ArgumentParser(description="SACAD v%s. Search and download an album cover." % (__version__),
                                       formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  arg_parser.add_argument("artist",
                          help="Artist to search for")
  arg_parser.add_argument("album",
                          help="Album to search for")
  arg_parser.add_argument("size",
                          type=int,
                          help="Target image size")
  arg_parser.add_argument("out_filepath",
                          help="Output image filepath")
  setup_common_args(arg_parser)
  arg_parser.add_argument("-v",
                          "--verbosity",
                          choices=("quiet", "warning", "normal", "debug"),
                          default="normal",
                          dest="verbosity",
                          help="Level of logging output")
  args = arg_parser.parse_args()
  args.format = os.path.splitext(args.out_filepath)[1][1:].lower()
  try:
    args.format = SUPPORTED_IMG_FORMATS[args.format]
  except KeyError:
    print("Unable to guess image format from extension, or unknown format: %s" % (args.format))
    exit(1)

  # setup logger
  logging_level = {"quiet": logging.CRITICAL + 1,
                   "warning": logging.WARNING,
                   "normal": logging.INFO,
                   "debug": logging.DEBUG}
  logging.getLogger().setLevel(logging_level[args.verbosity])
  if logging_level[args.verbosity] == logging.DEBUG:
    fmt = "%(threadName)s: %(message)s"
  else:
    fmt = "%(message)s"
  logging_formatter = colored_logging.ColoredFormatter(fmt=fmt)
  logging_handler = logging.StreamHandler()
  logging_handler.setFormatter(logging_formatter)
  logging.getLogger().addHandler(logging_handler)

  # search and download
  search_and_download(args.album,
                      args.artist,
                      args.format,
                      args.size,
                      args.size_tolerance_prct,
                      args.amazon_tlds,
                      args.no_lq_sources,
                      args.out_filepath)


if getattr(sys, "frozen", False):
  # fix for cx_freeze generated executable not finding certs
  os.environ["REQUESTS_CA_BUNDLE"] = os.path.join(os.path.dirname(sys.executable), "cacert.pem")


if __name__ == "__main__":
  cl_main()



#!/usr/bin/env python3

import collections
import contextlib
import functools
import os
import shutil
import tempfile
import unittest

import mutagen
import requests

import sacad.recurse as recurse


try:
  redirect_stdout = contextlib.redirect_stdout
except AttributeError:
  # contextlib.redirect_stdout is not available (Python 3.3), build our own
  import sys
  @contextlib.contextmanager
  def redirect_stdout(s):
    original_stdout, sys.stdout = sys.stdout, s
    try:
      yield
    finally:
      sys.stdout = original_stdout


class TestRecursive(unittest.TestCase):

  @classmethod
  def setUpClass(cls):
    cls.temp_dir = tempfile.TemporaryDirectory()
    cls.album1_dir = os.path.join(cls.temp_dir.name, "album1")
    os.mkdir(cls.album1_dir)
    url = "https://upload.wikimedia.org/wikipedia/en/4/45/ACDC_-_Back_In_Black-sample.ogg"
    with contextlib.closing(requests.get(url, stream=True)) as response:
      response.raise_for_status()
      filepath1 = os.path.join(cls.album1_dir, "2 track.ogg")
      with open(filepath1, "wb") as f:
        for chunk in response.iter_content(2 ** 14):
          f.write(chunk)
    mf = mutagen.File(filepath1)
    mf["artist"] = "ARTIST1"
    mf["album"] = "ALBUM1"
    mf.save()

    cls.album2_dir = os.path.join(cls.temp_dir.name, "album2")
    os.mkdir(cls.album2_dir)
    filepath2 = os.path.join(cls.album2_dir, "1.dat")
    with open(filepath2, "wb") as f:
      f.write(b"\x00" * 8)
    filepath3 = os.path.join(cls.album2_dir, "2 track.ogg")
    shutil.copyfile(filepath1, filepath3)
    mf = mutagen.File(filepath3)
    mf["artist"] = "ARTIST2"
    mf["album"] = "ALBUM2"
    mf.save()

    cls.not_album_dir = os.path.join(cls.temp_dir.name, "not an album")
    os.mkdir(cls.not_album_dir)
    shutil.copyfile(filepath2, os.path.join(cls.not_album_dir, "a.dat"))

    cls.invalid_album_dir = os.path.join(cls.temp_dir.name, "invalid album")
    os.mkdir(cls.invalid_album_dir)
    shutil.copyfile(filepath2, os.path.join(cls.invalid_album_dir, "2 track.ogg"))
    filepath4 = os.path.join(cls.invalid_album_dir, "3 track.ogg")
    shutil.copyfile(filepath1, filepath4)
    mf = mutagen.File(filepath4)
    del mf["album"]
    mf.save()

  @classmethod
  def tearDownClass(cls):
    cls.temp_dir.cleanup()

  def test_analyze_lib(self):
    with open(os.devnull, "wb") as dn, redirect_stdout(dn):
      work = recurse.analyze_lib(__class__.temp_dir.name, "a.jpg")
      self.assertEqual(len(work), 2)
      self.assertIn(__class__.album1_dir, work)
      self.assertEqual(work[__class__.album1_dir], ("ARTIST1", "ALBUM1"))
      self.assertIn(__class__.album2_dir, work)
      self.assertEqual(work[__class__.album2_dir], ("ARTIST2", "ALBUM2"))

      work = recurse.analyze_lib(__class__.temp_dir.name, "1.dat")
      self.assertEqual(len(work), 1)
      self.assertIn(__class__.album1_dir, work)
      self.assertEqual(work[__class__.album1_dir], ("ARTIST1", "ALBUM1"))

  def test_get_metadata(self):
    self.assertEqual(recurse.get_metadata(map(functools.partial(os.path.join,
                                                                __class__.album1_dir),
                                              os.listdir(__class__.album1_dir))),
                     ("ARTIST1", "ALBUM1"))
    self.assertEqual(recurse.get_metadata(map(functools.partial(os.path.join,
                                                                __class__.album2_dir),
                                              os.listdir(__class__.album2_dir))),
                     ("ARTIST2", "ALBUM2"))
    self.assertEqual(recurse.get_metadata(map(functools.partial(os.path.join,
                                                                __class__.not_album_dir),
                                              os.listdir(__class__.not_album_dir))),
                     (None, None))
    self.assertEqual(recurse.get_metadata(map(functools.partial(os.path.join,
                                                                __class__.invalid_album_dir),
                                              os.listdir(__class__.invalid_album_dir))),
                     ("ARTIST1", None))

  def test_analyze_dir(self):
    with open(os.devnull, "wb") as dn, redirect_stdout(dn):
      stats = collections.defaultdict(int)
      r = recurse.analyze_dir(stats,
                              __class__.album1_dir,
                              os.listdir(__class__.album1_dir),
                              "1.jpg",
                              None,
                              0)
      self.assertIn("files", stats)
      self.assertEqual(stats["files"], 1)
      self.assertIn("albums", stats)
      self.assertEqual(stats["albums"], 1)
      self.assertIn("missing covers", stats)
      self.assertEqual(stats["missing covers"], 1)
      self.assertNotIn("errors", stats)
      self.assertEqual(r[0], ("ARTIST1", "ALBUM1"))

      stats.clear()
      r = recurse.analyze_dir(stats,
                              __class__.album2_dir,
                              os.listdir(__class__.album2_dir),
                              "1.jpg",
                              None,
                              0)
      self.assertIn("files", stats)
      self.assertEqual(stats["files"], 2)
      self.assertIn("albums", stats)
      self.assertEqual(stats["albums"], 1)
      self.assertIn("missing covers", stats)
      self.assertEqual(stats["missing covers"], 1)
      self.assertNotIn("errors", stats)
      self.assertEqual(r[0], ("ARTIST2", "ALBUM2"))
      stats.clear()
      r = recurse.analyze_dir(stats,
                              __class__.album2_dir,
                              os.listdir(__class__.album2_dir),
                              "1.dat",
                              None,
                              0)
      self.assertIn("files", stats)
      self.assertEqual(stats["files"], 2)
      self.assertIn("albums", stats)
      self.assertEqual(stats["albums"], 1)
      self.assertNotIn("missing covers", stats)
      self.assertNotIn("errors", stats)
      self.assertEqual(r[0], (None, None))

      stats.clear()
      r = recurse.analyze_dir(stats,
                              __class__.not_album_dir,
                              os.listdir(__class__.not_album_dir),
                              "1.jpg",
                              None,
                              0)
      self.assertIn("files", stats)
      self.assertEqual(stats["files"], 1)
      self.assertNotIn("albums", stats)
      self.assertNotIn("missing covers", stats)
      self.assertNotIn("errors", stats)
      self.assertEqual(r[0], (None, None))

      stats.clear()
      r = recurse.analyze_dir(stats,
                              __class__.invalid_album_dir,
                              os.listdir(__class__.invalid_album_dir),
                              "1.jpg",
                              None,
                              0)
      self.assertIn("files", stats)
      self.assertEqual(stats["files"], 2)
      self.assertIn("albums", stats)
      self.assertEqual(stats["albums"], 1)
      self.assertIn("missing covers", stats)
      self.assertEqual(stats["missing covers"], 1)
      self.assertIn("errors", stats)
      self.assertEqual(stats["errors"], 1)
      self.assertEqual(r[0], ("ARTIST1", None))


if __name__ == "__main__":
  # run tests
  unittest.main()
